@@ -1,3 +1,4 @@
+#ident "$Id: automount.c,v 1.5 1998/03/29 10:26:20 hpa Exp $"
 /* ----------------------------------------------------------------------- *
  *
  *  automount.c - Linux automounter daemon
@@ -48,6 +49,7 @@ static char *pid_file = NULL;	        /* File in which to keep pid */
 int kproto_version;		        /* Kernel protocol version used */
 
 static int mounted = 0;
+static int submount = 0;
 
 struct pending_mount {
   pid_t pid;		/* Which process is mounting for us */
@@ -72,6 +74,9 @@ volatile struct pending_mount *junk_mounts = NULL;
 #define DEFAULT_TIMEOUT (5*60)	/* 5 minutes */
 #define CHECK_RATIO     4	/* exp_runfreq = exp_timeout/CHECK_RATIO */
 
+static void cleanup_exit(int exit_code);
+
+
 static int umount_ent(char *root, char *name)
 {
   char path_buf[PATH_MAX];
@@ -82,8 +87,8 @@ static int umount_ent(char *root, char *name)
   if ( !lstat(path_buf,&st) ) {
     if ( S_ISDIR(st.st_mode) ) {
       if ( st.st_dev != ap.dev ) {
-	rv = spawnl(LOG_NOTICE, _PATH_UMOUNT,
-		    _PATH_UMOUNT, path_buf, NULL);
+	rv = spawnl(LOG_DEBUG, PATH_UMOUNT,
+		    PATH_UMOUNT, path_buf, NULL);
 	if ( !rv )
 	  rmdir(path_buf);
       } else
@@ -131,28 +136,74 @@ static int umount_all(int force)
   return left;
 }	
 
-static int umount_autofs(int force)
+static int do_umount_autofs(void)
 {
   int rv;
   
-  if ( !mounted )
-    return -1;
-  
-  if ( umount_all(force) && !force )
-    return -1;
-  
   if (ap.ioctlfd >= 0) {
     ioctl(ap.ioctlfd, AUTOFS_IOC_CATATONIC, 0);
     close(ap.ioctlfd);
   }
   if (ap.pipefd >= 0)
     close(ap.pipefd);
-  rv = spawnl(LOG_ERR, _PATH_UMOUNT, _PATH_UMOUNT, ap.path, NULL);
+  rv = spawnl(LOG_ERR, PATH_UMOUNT, PATH_UMOUNT, ap.path, NULL);
+  if (rv == 0 && submount)
+    rmdir(ap.path);
+
   free(ap.path);
   mounted = 0;
   return rv;
 }
 
+static int umount_autofs(int force)
+{
+  if ( !mounted )
+    return -1;
+  if ( umount_all(force) && !force )
+    return -1;
+  return do_umount_autofs();
+}
+
+static void maybe_umount_autofs_and_exit(void)
+{
+  char path_buf[PATH_MAX];
+  struct stat st;
+  struct dirent *de;
+  DIR *dp;
+  int left;
+
+  /* If we are expiring, wait til next round to exit.  */
+  if ( ap.exp_process )
+    return;
+
+  chdir("/");
+  
+  dp = opendir(ap.path);
+  if ( !dp ) {
+    syslog(LOG_ERR, "maybe_umount_autofs: opendir: %m");
+    return;
+  }
+  
+  left = 0;
+  while ( (de = readdir(dp)) != NULL ) {
+    if ( strcmp(de->d_name,".") && strcmp(de->d_name,"..") ) {
+      sprintf(path_buf, "%s/%s", ap.path, de->d_name);
+      if ( !lstat(path_buf,&st) ) {
+        if ( S_ISDIR(st.st_mode) ) {
+          if ( st.st_dev != ap.dev )
+	    left++;
+	}
+      }
+    }
+  }
+  closedir(dp);
+
+  if (!left) {
+    if (do_umount_autofs() == 0)
+      cleanup_exit(0);
+  }
+}
+
 static int mount_autofs(char *path)
 {
   int pipefd[2];
@@ -189,7 +240,7 @@ static int mount_autofs(char *path)
 	  (unsigned)my_pgrp, AUTOFS_PROTO_VERSION);
   sprintf(our_name, "automount(pid%u)", (unsigned)my_pid);
   
-  if ( spawnl(LOG_CRIT, _PATH_MOUNT, _PATH_MOUNT, "-t", "autofs", "-o",
+  if ( spawnl(LOG_CRIT, PATH_MOUNT, PATH_MOUNT, "-t", "autofs", "-o",
 	      options, our_name, path, NULL) ) {
     close(pipefd[0]);
     close(pipefd[1]);
@@ -218,15 +269,20 @@ static int mount_autofs(char *path)
 
 static void run_expire(void)
 {
-  pid_t f;
   struct autofs_packet_expire pkt;
   sigset_t s, olds;
+  pid_t f;
   
   if ( ap.exp_process )	/* Another process is already expiring */
     return;		/* Try again next interval */
   
-  if ( ioctl(ap.ioctlfd, AUTOFS_IOC_EXPIRE, &pkt) )
-    return;		/* Don't fork if there is nothing do expire */
+  if ( ioctl(ap.ioctlfd, AUTOFS_IOC_EXPIRE, &pkt) ) {
+    /* Don't fork if there is nothing do expire, but if we are a submount,
+       see if maybe we shouldn't exit. */
+    if (submount)
+      maybe_umount_autofs_and_exit();
+    return;	
+  }
   
   /* Temporarily block SIGCHLD and SIGALRM between forking and setting
      ap.exp_process */
@@ -251,7 +307,7 @@ static void run_expire(void)
   close(ap.pipefd);
   
   syslog(LOG_DEBUG, "running expiration on path %s", ap.path);
-  
+
   do {
     if ( pkt.hdr.type == autofs_ptype_expire ) {
       if ( !umount_ent(ap.path,pkt.name) )
@@ -414,6 +470,8 @@ static void sig_child(int sig)
   while ( (pid = waitpid(-1, &status, WNOHANG)) > 0 ) {
     if ( pid == ap.exp_process ) {
       ap.exp_process = 0;
+      if (submount)
+	maybe_umount_autofs_and_exit();
     } else {
       for ( mtp = &ap.mounts ; (mt = *mtp) ; mtp = &mt->next ) {
 	if ( mt->pid == pid ) {
@@ -443,37 +501,36 @@ static void become_daemon(void)
   int nullfd;
   
   /* Don't BUSY any directories unneccessarily */
-
   chdir("/");
   
   /* Detach from foreground process */
-
-  pid = fork();
-  if ( pid > 0 )
-    exit(0);
-  else if ( pid < 0 ) {
-    fprintf(stderr, "%s: Could not detach process\n", program);
-    exit(1);
+  if ( !submount ) {
+    pid = fork();
+    if ( pid > 0 )
+      exit(0);
+    else if ( pid < 0 ) {
+      fprintf(stderr, "%s: Could not detach process\n", program);
+      exit(1);
+    }
   }
   
   /* Open syslog */ 
-
   openlog("automount", LOG_PID, LOG_DAEMON);
 
   /* Initialize global data */
-
   my_pid = getpid();
 
-  /* Make our own process group for "magic" reason */
-
-  if ( setpgrp() ) {
+  /* Make our own process group for "magic" reason: processes that share
+     our pgrp see the raw filesystem behine the magic.  So if we are a 
+     submount, don't change -- otherwise we won't be able to actually
+     perform the mount.  */
+  if ( !submount && setpgrp() ) {
     syslog(LOG_CRIT, "setpgrp: %m");
     exit(1);
   }
   my_pgrp = getpgrp();
-  
-  /* Redirect all our file descriptors to /dev/null */
 
+  /* Redirect all our file descriptors to /dev/null */
   if ( (nullfd = open("/dev/null", O_RDWR)) < 0 ) {
     syslog(LOG_CRIT, "cannot open /dev/null: %m");
     exit(1);
@@ -488,7 +545,6 @@ static void become_daemon(void)
   close(nullfd);
 
   /* Write pid file if requested */
-
   if ( pid_file ) {
     if ( (pidfp = fopen(pid_file, "wt")) ) {
       fprintf(pidfp, "%lu\n", (unsigned long) my_pid);
@@ -538,7 +594,8 @@ static void usage(void)
 
 int main(int argc, char *argv[])
 {
-  char *path, *map, *mapfmt, **mapargv;
+  char *path, *map, *mapfmt;
+  const char **mapargv;
   struct sigaction sa;
   int mapargc, opt;
   static const struct option long_options[] = {
@@ -546,6 +603,7 @@ int main(int argc, char *argv[])
     {"pid-file", 1, 0, 'p'},
     {"timeout",  1, 0, 't'},
     {"version",  0, 0, 'v'},
+    {"submount", 0, &submount, 1},
     {0,0,0,0}
   };
   
@@ -595,7 +653,7 @@ int main(int argc, char *argv[])
   
   path    = argv[0];
   map     = argv[1];
-  mapargv = &argv[2];
+  mapargv = (const char **) &argv[2];
   mapargc = argc-2;         
   
   syslog(LOG_INFO, "starting automounter version %s, path = %s, "
@@ -619,7 +677,7 @@ int main(int argc, char *argv[])
     syslog(LOG_CRIT, "%s: mount failed!", path);
     cleanup_exit(1);
   }	
-  
+
   /* The following signals cause a shutdown event to occur */
   sa.sa_handler = sig_shutdown;
   sigemptyset(&sa.sa_mask);
@@ -721,6 +779,11 @@ int main(int argc, char *argv[])
     if ( ap.exp_timeout )
       alarm(ap.exp_timeout + my_pid % ap.exp_runfreq);
   }
+
+  /* Initialization successful.  If we're a submount, send outself SIGSTOP to let
+     our parent know that we have grown up and don't need supervision anymore. */
+  if ( submount )
+    kill(my_pid, SIGSTOP);
   
   while ( !shutdown ) {
     if (handle_packet() && errno != EINTR)
