@@ -1,10 +1,10 @@
-#ident "$Id: mount.c,v 1.3 1998/03/27 04:15:48 hpa Exp $"
+#ident "$Id: mount.c,v 1.4 1999/12/17 19:02:47 hpa Exp $"
 /* ----------------------------------------------------------------------- *
  *   
  *  mount.c - Abstract mount code used by modules for an unexpected
  *            filesystem type
  *
- *   Copyright 1997 Transmeta Corporation - All Rights Reserved
+ *   Copyright 1997-2000 Transmeta Corporation - All Rights Reserved
  *
  *   This program is free software; you can redistribute it and/or modify
  *   it under the terms of the GNU General Public License as published by
@@ -24,8 +24,9 @@
 #include "automount.h"
 
 /* These filesystems are known not to work with the "generic" module */
-static char *not_generic[] = { "nfs", "smbfs", "ncpfs", "userfs", "afs",
-			       "autofs", NULL };
+/* Note: starting with Samba 2.0.6, smbfs is handled generically.    */
+static char *not_generic[] = { "nfs", "ncpfs", "userfs", "afs",
+			       "autofs", "changer", NULL };
 
 int do_mount(const char *root, const char *name, int name_len,
 	     const char *what, const char *fstype, const char *options)
